import copy
import os
from typing import List
from urllib.request import urlopen
from io import BytesIO
from zipfile import ZipFile

import numpy as np

from SRToolkit.dataset import SRDataset

from SRToolkit.utils import SymbolLibrary


class SRBenchmark:
    def __init__(self, benchmark_name: str, base_dir: str, metadata: dict = None):
        """
        Initializes an instance of the SRBenchmark class.

        Args:
            benchmark_name: The name of this benchmark.
            base_dir: The directory where the datasets will be stored.
            metadata: An optional dictionary containing metadata about this benchmark. This could include information such as the name of the benchmark, a citation for the benchmark, number of datasets, etc.
        """
        self.benchmark_name = benchmark_name
        self.base_dir = base_dir
        self.datasets = {}
        self.metadata = {} if metadata is None else metadata

    def add_dataset(self, dataset_name: str, ground_truth: List[str],  symbol_library: SymbolLibrary,
                    original_equation: str = None, max_evaluations: int=-1, max_expression_length: int=-1,
                    max_constants: int=8, success_threshold: float=1e-7, constant_range: List[float]=None,
                    num_variables: int=-1, dataset_metadata: dict=None):

        """
        Adds a dataset to the benchmark.

        Args:
            dataset_name: The name of the dataset.
            ground_truth: The ground truth expression, represented as a list of tokens (strings) in the infix notation.
            symbol_library: The symbol library to use.
            original_equation: The original equation from which the ground truth expression was generated.
            max_evaluations: The maximum number of expressions to evaluate. Less than 0 means no limit.
            max_expression_length: The maximum length of the expression. Less than 0 means no limit.
            max_constants: The maximum number of constants allowed in the expression. Less than 0 means no limit.
            success_threshold: The RMSE threshold below which the experiment is considered successful.
            constant_range: A list of two floats, specifying the lower and upper bounds for the constant values.
                Default is [-5.0, 5.0]. If constant_range is None, we automatically set it to [-5.0, 5.0]
                if the symbol library contains a symbol for constants.
            num_variables: The number of variables in the expression. Default is -1, which means we don't know.
            dataset_metadata: An optional dictionary containing metadata about this dataset. This could include
                information such as the name of the dataset, a citation for the dataset, number of variables, etc.
        """
        if original_equation is None:
            original_equation = "".join(ground_truth)

        self.datasets[dataset_name] = {
            "path": self.base_dir + "/" + dataset_name + ".npy",
            "ground_truth": ground_truth,
            "original_equation": original_equation,
            "symbols": symbol_library,
            "max_evaluations": max_evaluations,
            "max_expression_length": max_expression_length,
            "max_constants": max_constants,
            "success_threshold": success_threshold,
            "constant_range": constant_range,
            "dataset_metadata": self.metadata.update(dataset_metadata),
            "num_variables": num_variables
        }

    def create_dataset(self, dataset_name: str):
        """
        Creates an instance of a dataset from the given dataset name.

        Args:
            dataset_name: The name of the dataset to create.

        Returns:
            A SRDataset instance containing the data, ground truth expression, and metadata for the given dataset.

        Raises:
            ValueError: If the dataset name is not found in the available datasets.
        """
        if dataset_name in self.datasets:
            # Check if dataset exists otherwise download it from an url
            if os.path.exists(self.datasets[dataset_name]["path"]):
                data = np.load(self.datasets[dataset_name]["path"])
            else:
                raise ValueError(f"Could not find dataset {dataset_name} at {self.datasets[dataset_name]['path']}")

            X = data[:, :-1]
            y = data[:, -1]

            return SRDataset(X, y, ground_truth=self.datasets[dataset_name]["ground_truth"],
                             original_equation=self.datasets[dataset_name]["original_equation"],
                             symbols=self.datasets[dataset_name]["symbols"],
                             max_evaluations=self.datasets[dataset_name]["max_evaluations"],
                             max_expression_length=self.datasets[dataset_name]["max_expression_length"],
                             max_constants=self.datasets[dataset_name]["max_constants"],
                             success_threshold=self.datasets[dataset_name]["success_threshold"],
                             constant_range=self.datasets[dataset_name]["constant_range"],
                             dataset_metadata=self.datasets[dataset_name]["dataset_metadata"])
        else:
            raise ValueError(f"Dataset {dataset_name} not found")

    def list_datasets(self, verbose=True, num_variables: int=-1):
        """
        Lists the available datasets.

        Args:
            verbose (bool): If True, also prints out a description of each dataset.
            num_variables (int): If not -1, only show datasets with the given number of variables.

        Returns:
            A list of dataset names.
        """
        datasets = [dataset_name for dataset_name in self.datasets if num_variables < 0 or self.datasets[dataset_name]["num_variables"] == num_variables]
        datasets = sorted(datasets, key= lambda dataset_name: (self.datasets[dataset_name]["num_variables"], dataset_name))

        if verbose:
            # TODO: Make all names be of equal length for nicer output
            for d in datasets:
                if self.datasets[d]["num_variables"] == 1:
                    variable_str = "1 variable"
                elif self.datasets[d]["num_variables"] < 1:
                    variable_str = "Amount of variables unknown"
                else:
                    variable_str = f"{self.datasets[d]['num_variables']} variables"

                print(f"{d}:\t{variable_str}, \tExpression: {self.datasets[d]['original_equation']}")
        return datasets


    @staticmethod
    def download_benchmark_data(url, directory_path):
        # Check if directory_path exist
        """
        Downloads a benchmark dataset from the given url to the given directory path.

        This function will first check if the directory_path exists. If not, it will create it. Then it will check if the directory_path is empty. If it is not empty, it will not download the data. If it is empty, it will download the data from the given url and extract it to the directory_path.

        Args:
            url (str): The url of the benchmark dataset to download.
            directory_path (str): The path of the directory where the dataset should be downloaded.

        Returns:
            None
        """
        if not os.path.exists(directory_path):
            os.makedirs(directory_path)

        # Check if directory_path is empty
        if not os.listdir(directory_path):
            # Download data from the url to the directory_path
            http_response = urlopen(url)
            zipfile = ZipFile(BytesIO(http_response.read()))
            zipfile.extractall(path=directory_path)


    @staticmethod
    def feynman(dataset_directory: str):
        """
        Downloads the Feynman benchmark dataset, sets up symbol libraries, and adds predefined datasets to the benchmark.

        This method downloads the Feynman benchmark dataset from a specified URL, initializes symbol libraries for symbolic regression with varying numbers of variables, and adds multiple predefined datasets to the benchmark with their respective equations and metadata.

        Args:
            dataset_directory (str): The directory path where the benchmark dataset will be downloaded and stored.

        Returns:
            SRBenchmark: An instance of the SRBenchmark class containing the predefined datasets.
        """
        url = "https://raw.githubusercontent.com/smeznar/SymbolicRegressionToolkit/master/data/feynman.zip"
        SRBenchmark.download_benchmark_data(url, dataset_directory)

        sl_1v = SymbolLibrary()
        sl_1v.add_symbol("+", symbol_type="op", precedence=0, np_fn="{} = {} + {}")
        sl_1v.add_symbol("-", symbol_type="op", precedence=0, np_fn="{} = {} - {}")
        sl_1v.add_symbol("*", symbol_type="op", precedence=1, np_fn="{} = {} * {}")
        sl_1v.add_symbol("/", symbol_type="op", precedence=1, np_fn="{} = {} / {}")
        sl_1v.add_symbol("sin", symbol_type="fn", precedence=5, np_fn="{} = np.sin({})")
        sl_1v.add_symbol("cos", symbol_type="fn", precedence=5, np_fn="{} = np.cos({})")
        sl_1v.add_symbol("exp", symbol_type="fn", precedence=5, np_fn="{} = np.exp({})")
        sl_1v.add_symbol("sqrt", symbol_type="fn", precedence=5, np_fn="{} = np.sqrt({})")
        sl_1v.add_symbol("^2", symbol_type="fn", precedence=5, np_fn="{} = np.power({}, 2)")
        sl_1v.add_symbol("^3", symbol_type="fn", precedence=5, np_fn="{} = np.power({}, 3)")
        sl_1v.add_symbol("C", "const", 5, "C[{}]")
        sl_1v.add_symbol("X_0", "var", 5, "X[:, 0]")

        sl_2v = copy.copy(sl_1v)
        sl_2v.add_symbol("X_1", "var", 5, "X[:, 1]")

        sl_3v = copy.copy(sl_2v)
        sl_3v.add_symbol("X_2", "var", 5, "X[:, 2]")

        sl_4v = copy.copy(sl_3v)
        sl_4v.add_symbol("X_3", "var", 5, "X[:, 3]")

        sl_5v = copy.copy(sl_4v)
        sl_5v.add_symbol("X_4", "var", 5, "X[:, 4]")

        sl_6v = copy.copy(sl_5v)
        sl_6v.add_symbol("X_5", "var", 5, "X[:, 5]")

        sl_8v = copy.copy(sl_6v)
        sl_8v.add_symbol("X_6", "var", 5, "X[:, 6]")
        sl_8v.add_symbol("X_7", "var", 5, "X[:, 7]")

        sl_9v = copy.copy(sl_8v)
        sl_9v.add_symbol("X_8", "var", 5, "X[:, 8]")

        benchmark = SRBenchmark("feynman", dataset_directory)
        benchmark.add_dataset("I.16.6", ["(", "X_2", "+", "X_1", ")", "/", "(", "1", "+", "X_2", "*", "X_1", "/", "X_1", "^2",")"], sl_3v, original_equation="v1 = (u+v)/(1+u*v/c^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.15.4", ["(", "0", "-", "X_0", ")", "*", "X_1", "*", "cos", "(", "X_2", ")"], sl_3v, original_equation="E_n = -mom*B*cos(theta)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.27.16", ["X_0", "*", "X_1", "*", "X_2", "^2"], sl_3v, original_equation="flux = epsilon*c*Ef^2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.11.19", ["X_0", "*", "X_3", "+", "X_1", "*", "X_4", "+", "X_2", "*", "X_5"], sl_6v, original_equation="A = x1*y1+x2*y2+x3*y3", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=6,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.15.3x", ["(", "X_0", "-", "X_1", "*", "X_3", ")", "/", "sqrt", "(", "1", "-", "X_1", "^2", "/", "X_2", "^2", ")"], sl_4v, original_equation="x1 = (x-u*t)/sqrt(1-u^2/c^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.10.7", ["X_0", "/", "sqrt", "(", "1", "-", "X_1", "^2", "/", "X_2", "^2", ")"], sl_3v, original_equation="m = m_0/sqrt(1-v^2/c^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.9.18", ["X_2", "*", "X_0", "*", "X_1", "/", "(", "(", "X_4", "-", "X_3", ")", "^2", "+", "(", "X_6", "-", "X_5", ")", "^2", "+", "(", "X_8", "-", "X_7", ")", "^2",")"], sl_9v, original_equation="F = G*m1*m2/((x2-x1)^2+(y2-y1)^2+(z2-z1)^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=9,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.15.3t", ["(", "X_3", "-", "X_2", "*", "X_0", "/", "X_1", "^2", ")", "/", "sqrt", "(", "1", "-", "X_2", "^2", "/", "X_1", "^2", ")"], sl_4v, original_equation="t1 = (t-u*x/c^2)/sqrt(1-u^2/c^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.36.38", ["X_0", "*", "X_1", "/", "(", "X_2", "*", "X_3",")", "+", "(", "X_0", "*", "X_4", ")", "/", "(", "X_5", "*", "X_6", "^2", "*", "X_2", "*", "X_3" ")", "*", "X_7"], sl_8v, original_equation="f = mom*H/(kb*T)+(mom*alpha)/(epsilon*c**2*kb*T)*M", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=8,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.43.43", ["1", "/", "(", "X_0", "-", "1", ")", "*", "X_1", "*", "X_3", "/", "X_2"], sl_4v, original_equation="kappa = 1/(gamma-1)*kb*v/A", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.15.5", ["(", "0", "-", "X_0", ")", "*", "X_1", "*", "cos", "(", "X_2", ")"], sl_3v, original_equation="E_n = -p_d*Ef*cos(theta)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.37.4", ["X_0", "+", "X_1", "+", "2", "*", "sqrt", "(", "X_0", "*", "X_1", ")", "*", "cos", "(", "X_2", ")"], sl_3v, original_equation="Int = I1+I2+2*sqrt(I1*I2)*cos(delta)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.6.11", ["1", "/", "(", "4", "*", "pi", "*", "X_0", ")", "*", "X_1", "*", "cos", "(", "X_2", ")", "/", "X_3", "^2"], sl_4v, original_equation="Volt = 1/(4*pi*epsilon)*p_d*cos(theta)/r^2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.7.38", ["2", "*", "X_0", "*", "X_1", "/", "(", "X_2", "/", "(", "2", "*", "pi", ")", ")"], sl_3v, original_equation="omega = 2*mom*B/(h/(2*pi))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.34.2a", ["X_0", "*", "X_1", "/", "(", "2", "*", "pi", "*", "X_2", ")"], sl_3v, original_equation="l = q*v/(2*pi*r)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.13.23", ["X_0", "/", "sqrt", "(", "1", "-", "X_1", "^2", "/", "X_2", "^2", ")"], sl_3v, original_equation="rho_c = rho_c_0/sqrt(1-v^2/c^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.29.4", ["X_0", "/", "X_1"], sl_2v, original_equation="k = omega/c", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.38.12", ["4", "*", "pi", "*", "X_3", "*", "(", "X_2", "/", "(", "2", "*", "pi", ")", ")", "^2", "/", "(", "X_0", "*", "X_1", "^2", ")"], sl_4v, original_equation="r = 4*pi*epsilon*(h/(2*pi))^2/(m*q^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.15.27", ["2", "*", "pi", "*", "X_0", "/", "(", "X_1", "*", "X_2", ")"], sl_3v, original_equation="k = 2*pi*alpha/(n*d)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.41.16", ["X_2", "/", "(", "2", "*", "pi", ")", "*", "X_0", "^3", "/", "(", "pi", "^2", "*", "X_4", "^2", "*", "(", "exp", "(", "(", "X_2", "/", "(", "2", "*", "pi", ")", ")", "*", "X_0", "/", "(", "X_3", "*", "X_1", ")", ")", "-", "1", ")", ")"], sl_5v, original_equation="L_rad = h/(2*pi)*omega^3/(pi^2*c^2*(exp((h/(2*pi))*omega/(kb*T))-1))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.48.20", ["X_0", "*", "X_2", "^2", "/", "sqrt", "(", "1", "-", "X_1", "^2", "/", "X_2", "^2", ")"], sl_3v, original_equation="E_n = m*c^2/sqrt(1-v^2/c^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.11.20", ["X_0", "*", "X_1", "^2", "*", "X_2", "/", "(", "3", "*", "X_3", "*", "X_4", ")"], sl_5v, original_equation="Pol = n_rho*p_d^2*Ef/(3*kb*T)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.25.13", ["X_0", "/", "X_1"], sl_2v, original_equation="Volt = q/C", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.15.12", ["2", "*", "X_0", "*", "cos", "(", "X_1", "*", "X_2", ")"], sl_3v, original_equation="E_n = 2*U*(1-cos(k*d))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.24.6", ["0.25", "*", "X_0", "*", "(", "X_1", "^2", "+", "X_2", "^2", ")", "*", "X_3", "^2"], sl_4v, original_equation="E_n = 1/2*m*(omega^2+omega_0^2)*1/2*x^2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.34.27", ["(", "X_1", "/", "(", "2", "*", "pi", ")", ")", "*", "X_0"], sl_2v, original_equation="E_n =(h/(2*pi))*omega", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.43.31", ["X_0", "*", "X_2", "*", "X_1"], sl_3v, original_equation="D = mob*kb*T", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.29.16", ["sqrt", "(", "X_0", "^2", "+", "X_1", "^2", "-", "2", "*", "X_0", "*", "X_1", "*", "cos", "(", "X_2", "-", "X_3", ")", ")"], sl_4v, original_equation="x = sqrt(x1^2+x2^2-2*x1*x2*cos(theta1-theta2))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.18.4", ["(","X_0", "*", "X_2", "+", "X_1", "*", "X_3", ")", "/", "(", "X_0", "+", "X_1", ")"], sl_4v, original_equation="r = (m1*r1+m2*r2)/(m1+m2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.6.15a", ["X_1", "*", "(", "4", "*", "pi", "*", "X_0", ")", "*", "3", "*", "X_5", "/", "(", "X_2", "^2", "*", "X_2", "^3", ")", "*", "sqrt", "(", "X_3", "^2", "+", "X_4", "^2", ")"], sl_6v, original_equation="Ef = p_d/(4*pi*epsilon)*3*z/r^5*sqrt(x^2+y^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=6,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.30.3", ["X_0", "*", "sin", "(", "X_2", "*", "X_1", "/", "2", ")", "^2", "/", "sin", "(", "X_1", "/", "2", ")", "^2"], sl_3v, original_equation="Int = Int_0*sin(n*theta/2)^2/sin(theta/2)^2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.9.52", ["(", "X_0", "*", "X_1", "*", "X_2", "/", "(", "X_3", "/", "(", "2", "*", "pi", ")", ")", ")", "*", "sin", "(", "(", "X_4", "-", "X_5", ")", "*", "X_2", "/", "2", ")", "^2", "/", "(", "(", "X_4", "-", "X_5", ")", "*", "X_2", "/", "2", ")","^2"], sl_6v, original_equation="prob = (p_d*Ef*t/(h/(2*pi)))*sin((omega-omega_0)*t/2)^2/((omega-omega_0)*t/2)^2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=6,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.34.2", ["X_0", "*", "X_1", "*", "X_2", "/", "2"], sl_3v, original_equation="mom = q*v*r/2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.39.11", ["1", "/", "(", "X_0", "-", "1", ")", "*", "X_1", "*", "X_2"], sl_3v, original_equation="E_n = 1/(gamma-1)*pr*V", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.11.28", ["1", "+", "X_0", "*", "X_1", "/", "(", "1", "-", "(", "X_0", "*", "X_1", "/", "3", ")", ")"], sl_2v, original_equation="theta = 1+n*alpha/(1-(n*alpha/3))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.3.24", ["X_0", "/", "(", "4", "*", "pi", "*", "X_1", "^2", ")"], sl_2v, original_equation="flux = Pwr/(4*pi*r^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.24.17", ["sqrt", "(", "X_0", "^2", "/", "X_1", "^2", "-", "pi", "^2", "/", "X_2", "^2", ")"], sl_3v, original_equation="k = sqrt(omega^2/c^2-pi^2/d^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.13.17", ["1", "/", "(", "4", "*", "pi", "*", "X_0", "*", "X_1", "^2", ")", "*", "2", "*", "X_2", "/", "X_3"], sl_4v, original_equation="B = 1/(4*pi*epsilon*c^2)*2*I/r", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.12.5", ["X_0", "*", "X_1"], sl_2v, original_equation="F = q2*Ef", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.35.18", [], sl_5v, original_equation="n_0/(exp(mom*B/(kb*T))+exp(-mom*B/(kb*T)))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.34.11", [], sl_4v, original_equation="g_*q*B/(2*m)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.34.29a", [], sl_3v, original_equation="q*h/(4*pi*m)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.32.17", [], sl_6v, original_equation="(1/2*epsilon*c*Ef**2)*(8*pi*r**2/3)*(omega**4/(omega**2-omega_0**2)**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=6,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.35.21", [], sl_5v, original_equation="n_rho*mom*tanh(mom*B/(kb*T))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.44.4", [], sl_5v, original_equation="n*kb*T*ln(V2/V1)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.4.32", [], sl_4v, original_equation="1/(exp((h/(2*pi))*omega/(kb*T))-1)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.10.9", [], sl_3v, original_equation="sigma_den/epsilon*1/(1+chi)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.38.3", [], sl_4v, original_equation="Y*A*x/d", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.6.2b", [], sl_3v, original_equation="exp(-((theta-theta1)/sigma)**2/2)/(sqrt(2*pi)*sigma)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.8.31", [], sl_2v, original_equation="epsilon*Ef**2/2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.6.2a", [], sl_1v, original_equation="exp(-theta**2/2)/sqrt(2*pi)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.12.43", [], sl_2v, original_equation="n*(h/(2*pi))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.17.37", [], sl_3v, original_equation="beta*(1+alpha*cos(theta))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.10.19", [], sl_4v, original_equation="mom*sqrt(Bx**2+By**2+Bz**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.11.7", [], sl_6v, original_equation="n_0*(1+p_d*Ef*cos(theta)/(kb*T))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=6,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.39.1", [], sl_2v, original_equation="3/2*pr*V", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.37.1", [], sl_3v, original_equation="mom*(1+chi)*B", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.12.4", [], sl_3v, original_equation="q1*r/(4*pi*epsilon*r**3)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.27.18", [], sl_2v, original_equation="epsilon*Ef**2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.12.2", [], sl_4v, original_equation="q1*q2*r/(4*pi*epsilon*r**3)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.13.18", [], sl_4v, original_equation="2*E_n*d**2*k/(h/(2*pi))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.11.3", [], sl_5v, original_equation="q*Ef/(m*(omega_0**2-omega**2))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.40.1", [], sl_6v, original_equation="n_0*exp(-m*g*x/(kb*T))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=6,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.21.20", [], sl_4v, original_equation="-rho_c_0*q*A_vec/m", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.43.16", [], sl_4v, original_equation="mu_drift*q*Volt/d", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.15.10", [], sl_3v, original_equation="m_0*v/sqrt(1-v**2/c**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.30.5", [], sl_3v, original_equation="arcsin(lambd/(n*d))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.50.26", [], sl_4v, original_equation="x1*(cos(omega*t)+alpha*cos(omega*t)**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.12.11", [], sl_5v, original_equation="q*(Ef+B*v*sin(theta))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.6.2", [], sl_2v, original_equation="exp(-(theta/sigma)**2/2)/(sqrt(2*pi)*sigma)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.14.4", [], sl_2v, original_equation="1/2*k_spring*x**2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.47.23", [], sl_3v, original_equation="sqrt(gamma*pr/rho)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.8.7", [], sl_3v, original_equation="3/5*q**2/(4*pi*epsilon*d)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.15.14", [], sl_3v, original_equation="(h/(2*pi))**2/(2*E_n*d**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.34.14", [], sl_3v, original_equation="(1+v/c)/sqrt(1-v**2/c**2)*omega_0", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.8.54", [], sl_3v, original_equation="sin(E_n*t/(h/(2*pi)))**2", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.26.2", [], sl_2v, original_equation="arcsin(n*sin(theta2))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.19.51", [], sl_5v, original_equation="-m*q**4/(2*(4*pi*epsilon)**2*(h/(2*pi))**2)*(1/n**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.4.33", [], sl_4v, original_equation="(h/(2*pi))*omega/(exp((h/(2*pi))*omega/(kb*T))-1)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.34.1", [], sl_3v, original_equation="omega_0/(1-v/c)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.11.27", [], sl_4v, original_equation="n*alpha/(1-(n*alpha/3))*epsilon*Ef", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.13.34", [], sl_3v, original_equation="rho_c_0*v/sqrt(1-v**2/c**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.4.23", [], sl_3v, original_equation="q/(4*pi*epsilon*r)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.32.5", [], sl_4v, original_equation="q**2*a**2/(6*pi*epsilon*c**3)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.13.12", [], sl_5v, original_equation="G*m1*m2*(1/r2-1/r1)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.2.42", [], sl_5v, original_equation="kappa*(T2-T1)*A/d", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.27.6", [], sl_3v, original_equation="1/(1/d1+n/d2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("III.14.14", [], sl_5v, original_equation="I_0*(exp(q*Volt/(kb*T))-1)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.18.12", [], sl_3v, original_equation="r*F*sin(theta)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.18.14", [], sl_4v, original_equation="m*r*v*sin(theta)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.21.32", [], sl_5v, original_equation="q/(4*pi*epsilon*r*(1-v/c))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.38.14", [], sl_2v, original_equation="Y/(2*(1+sigma))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.34.8", [], sl_4v, original_equation="q*v*B/p", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.8.14", [], sl_4v, original_equation="sqrt((x2-x1)**2+(y2-y1)**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.6.15b", [], sl_4v, original_equation="p_d/(4*pi*epsilon)*3*cos(theta)*sin(theta)/r**3", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.12.1", [], sl_2v, original_equation="mu*Nn", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("II.34.29b", [], sl_5v, original_equation="g_*mom*B*Jz/(h/(2*pi))", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=5,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.13.4", [], sl_4v, original_equation="1/2*m*(v**2+u**2+w**2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.39.22", [], sl_4v, original_equation="n*kb*T/V", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=4,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])
        benchmark.add_dataset("I.14.3", [], sl_3v, original_equation="m*g*z", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=3,
                              dataset_metadata=benchmark.metadata, constant_range=[-5.0, 5.0])

        return benchmark

    @staticmethod
    def nguyen(dataset_directory: str):
        """
        Downloads and initializes the Nguyen benchmark datasets for symbolic regression.

        This method downloads the Nguyen symbolic regression benchmark datasets from a specified URL
        and initializes a set of datasets using a provided dataset directory. It creates two symbol libraries
        for equations with one variable and two variables, respectively, and populates the benchmark with various
        Nguyen equations, each represented with its symbolic tokens and associated symbol library.

        Args:
            dataset_directory (str): The directory where the benchmark datasets will be stored and accessed.

        Returns:
            SRBenchmark: An initialized SRBenchmark instance containing the Nguyen datasets.
        """
        url = "https://raw.githubusercontent.com/smeznar/SymbolicRegressionToolkit/master/data/nguyen.zip"
        SRBenchmark.download_benchmark_data(url, dataset_directory)
        # we create a SymbolLibrary with 1 and with 2 variables
        # Each library contains +, -, *, /, sin, cos, exp, log, sqrt, ^2, ^3
        sl_1v = SymbolLibrary()
        sl_1v.add_symbol("+", symbol_type="op", precedence=0, np_fn="{} = {} + {}")
        sl_1v.add_symbol("-", symbol_type="op", precedence=0, np_fn="{} = {} - {}")
        sl_1v.add_symbol("*", symbol_type="op", precedence=1, np_fn="{} = {} * {}")
        sl_1v.add_symbol("/", symbol_type="op", precedence=1, np_fn="{} = {} / {}")
        sl_1v.add_symbol("sin", symbol_type="fn", precedence=5, np_fn="{} = np.sin({})")
        sl_1v.add_symbol("cos", symbol_type="fn", precedence=5, np_fn="{} = np.cos({})")
        sl_1v.add_symbol("exp", symbol_type="fn", precedence=5, np_fn="{} = np.exp({})")
        sl_1v.add_symbol("log", symbol_type="fn", precedence=5, np_fn="{} = np.log10({})")
        sl_1v.add_symbol("sqrt", symbol_type="fn", precedence=5, np_fn="{} = np.sqrt({})")
        sl_1v.add_symbol("^2", symbol_type="fn", precedence=5, np_fn="{} = np.power({}, 2)")
        sl_1v.add_symbol("^3", symbol_type="fn", precedence=5, np_fn="{} = np.power({}, 3)")
        sl_1v.add_symbol("X_0", "var", 5, "X[:, 0]")

        sl_2v = copy.copy(sl_1v)
        sl_2v.add_symbol("X_1", "var", 5, "X[:, 1]")

        # Add datasets to the benchmark
        benchmark = SRBenchmark("Nguyen", dataset_directory)
        benchmark.add_dataset("NG-1", ["X_0", "+", "X_0", "^2", "+", "X_0", "^3"], sl_1v,
                              original_equation="x+x^2+x^3", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-2", ["X_0", "+", "X_0", "^2", "+", "X_0", "^3", "+", "X_0","*", "X_0", "^3"], sl_1v,
                              original_equation="x+x^2+x^3+x^4", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-3", ["X_0", "+", "X_0", "^2", "+", "X_0", "^3", "+", "X_0","*", "X_0", "^3", "+", "X_0","^2", "*", "X_0", "^3"], sl_1v,
                              original_equation="x+x^2+x^3+x^4+x^5", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-4", ["X_0", "+", "X_0", "^2", "+", "X_0", "^3", "+", "X_0","*", "X_0", "^3", "+", "X_0","^2", "*", "X_0", "^3", "+", "X_0","^3", "*", "X_0", "^3"], sl_1v,
                              original_equation="x+x^2+x^3+x^4+x^5+x^6", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-5", ["sin", "(", "X_0", "^2", ")", "*", "cos", "(", "X_0", ")", "-", "1"], sl_1v,
                              original_equation="sin(x^2)*cos(x)-1", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-6", ["sin", "(", "X_0", ")", "+", "sin", "(", "X_0", "+", "X_0", "^2", ")"], sl_1v,
                              original_equation="sin(x)+sin(x+x^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-7", ["log", "(", "1", "+", "X_0", ")", "+", "log", "(", "1", "+", "X_0", "^2", ")"], sl_1v,
                              original_equation="log(1+x)+log(1+x^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-8", ["sqrt", "(", "X_0", ")"], sl_1v,
                              original_equation="sqrt(x)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=1,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-9", ["sin", "(", "X_0", ")", "+", "sin", "(", "X_1", "^2", ")"], sl_2v,
                              original_equation="sin(x)+sin(y^2)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata)
        benchmark.add_dataset("NG-10", ["2", "*", "sin", "(", "X_0", ")", "*", "cos", "(", "X_1", ")"], sl_2v,
                              original_equation="2*sin(x)*cos(y)", max_evaluations=100000,
                              max_expression_length=50, success_threshold=1e-7, num_variables=2,
                              dataset_metadata=benchmark.metadata)

        return benchmark



from typing import List

import numpy as np

from SRToolkit.evaluation import SR_evaluator
from SRToolkit.utils import SymbolLibrary

class SRDataset:
    def __init__(self, X: np.ndarray, y: np.ndarray, ground_truth: List[str], original_equation: str,
                 symbols: SymbolLibrary, max_evaluations: int=-1, max_expression_length: int=-1, max_constants: int=8,
                 success_threshold: float=1e-7, constant_range: List[float]=None, dataset_metadata: dict=None):
        """
        Initializes an instance of the SRDataset class.

        Args:
            X: The input data to be used in parameter estimation for variables. We assume that X is a 2D array
                with shape (n_samples, n_features).
            y: The target values to be used in parameter estimation.
            ground_truth: The ground truth expression, represented as a list of tokens (strings) in the infix notation.
            original_equation: The original equation from which the ground truth expression was generated).
            symbols: The symbol library to use.
            max_evaluations: The maximum number of expressions to evaluate. Less than 0 means no limit.
            max_expression_length: The maximum length of the expression. Less than 0 means no limit.
            max_constants: The maximum number of constants allowed in the expression. Less than 0 means no limit.
            success_threshold: The RMSE threshold below which the experiment is considered successful.
            constant_range: A list of two floats, specifying the lower and upper bounds for the constant values.
                Default is [-5.0, 5.0]. If constant_range is None, we automatically set it to [-5.0, 5.0]
                if the symbol library contains a symbol for constants.
            dataset_metadata: An optional dictionary containing metadata about this evaluation. This could include
                information such as the name of the dataset, a citation for the dataset, number of variables, etc.
        """
        self.X = X
        self.y = y
        self.ground_truth = "".join(ground_truth)
        self.original_equation = original_equation

        self.max_evaluations = max_evaluations
        self.max_expression_length = max_expression_length
        self.max_constants = max_constants

        self.success_threshold = success_threshold

        self.symbols = symbols

        # See if symbols contain a symbol for constants
        symbols_metadata = self.symbols.symbols.values()
        self.contains_constants = any([symbol["type"] == "const" for symbol in symbols_metadata])
        if constant_range is None and self.contains_constants:
            constant_range = [-5.0, 5.0]
        self.constant_range = constant_range

        self.dataset_metadata = dataset_metadata

    def __str__(self) -> str:
        """
        Returns a string describing this dataset.

        The string describes the target expression, symbols that should be used,
        and the success threshold. It also includes any constraints that should
        be followed when evaluating a model on this dataset, such as the maximum
        number of expressions to evaluate, the maximum length of the expression,
        and the maximum number of constants allowed in the expression. If the
        symbol library contains a symbol for constants, the string also includes
        the range of constants.

        For other metadata, please refer to the attribute self.dataset_metadata.

        Returns:
            A string describing this dataset.
        """
        description = f"Dataset for target expression {self.ground_truth}. "
        description += (f" When evaluating your model on this dataset, you should limit your generative model to only"
                        f" produce expressions using the following symbols: {str(self.symbols)}. Expressions are deemed"
                        f" successful if the root mean squared error is less than {self.success_threshold}. However, we"
                        f" advise that you check the best performing expressions manually to ensure they are correct.")

        has_limitations = False
        limitations = "Constraints for this dataset are:"
        if self.max_evaluations > 0:
            has_limitations = True
            limitations += f" max_evaluations={self.max_evaluations}, "
        if self.max_expression_length > 0:
            has_limitations = True
            limitations += f" max_expression_length={self.max_expression_length}, "
        if self.max_constants > 0:
            has_limitations = True
            limitations += f" max_constants={self.max_constants}, "

        if has_limitations:
            limitations = limitations[:-2] + "."
            description += limitations

        if self.contains_constants:
            description += f" The dataset contains constants. The range of constants is {self.constant_range}."

        description += "For other metadata, please refer to the attribute self.dataset_metadata."

        return description

    def create_evaluator(self, metadata: dict=None) -> SR_evaluator:
        """
        Creates an instance of the SR_evaluator class from this dataset.

        Args:
            metadata: An optional dictionary containing metadata about this evaluation. This could include
                information such as the dataset used, the model used, seed, etc.

        Returns:
            An instance of the SR_evaluator class.
        """
        if metadata is None:
            metadata = dict()
        metadata["dataset_metadata"] = self.dataset_metadata
        return SR_evaluator(self.X, self.y, max_evaluations=self.max_evaluations, metadata=metadata,
                            symbol_library=self.symbols, max_constants=self.max_constants,
                            max_expression_length=self.max_expression_length,)


"""
Symbolic Regression Toolkit

This package provides a Python-based toolkit for equation discovery/symbolic regression.

Modules:
    utils: The module containing utility classes and functions.
    evaluation: The module containing classes and functions for estimating parameters and evaluating Symbolic Regression models.
"""

__version__ = "1.2.2"
__author__ = "Sebastian Mežnar"
__credits__ = "Jožef Stefan Institute"

from . import utils
from . import evaluation
from . import dataset


"""
This module contains the SymbolLibrary class, which is used for managing symbols and their properties.
"""
import copy

class SymbolLibrary:
    def __init__(self):
        """
        Initializes an instance of the SymbolLibrary class. This class is used for managing symbols and their
        properties for other functionality in this package.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.get_type("x")
            'var'
            >>> library.get_precedence("x")
            0
            >>> library.get_np_fn("x")
            'x'
            >>> library.remove_symbol("x")
            >>> library = SymbolLibrary.default_symbols()

        Attributes:
            symbols : dict
                A dictionary mapping symbols to their properties (type, precedence, numpy function).

        Methods:
            add_symbol(symbol, symbol_type, precedence, np_fn):
                Adds a symbol to the library.
            remove_symbol(symbol):
                Removes a symbol from the library.
            get_type(symbol):
                Retrieves the type of a symbol from the library.
            get_precedence(symbol):
                Returns the precedence of the given symbol.
            get_np_fn(symbol):
                Returns the numpy function corresponding to the given symbol.
            default_symbols():
                Returns a SymbolLibrary with the default symbols.
        """
        self.symbols = dict()

    def __str__(self) -> str:
        """
        Returns a string representation of the SymbolLibrary instance.

        This method provides a comma-separated string of all the symbol keys
        currently stored in the SymbolLibrary.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> str(library)
            'x'
            >>> library.add_symbol("sin", "fn", 5, "{} = np.sin({})")
            >>> str(library)
            'x, sin'

        Returns:
            A string containing all symbols in the library, separated by commas.
        """
        return ", ".join(self.symbols.keys())

    def __copy__(self) -> "SymbolLibrary":
        """
        Creates a copy of the SymbolLibrary instance.

        Examples:
            >>> old_symbols = SymbolLibrary()
            >>> old_symbols.add_symbol("x", "var", 0, "x")
            >>> print(old_symbols)
            x
            >>> new_symbols = copy.copy(old_symbols)
            >>> new_symbols.add_symbol("sin", "fn", 5, "{} = np.sin({})")
            >>> print(old_symbols)
            x
            >>> print(new_symbols)
            x, sin

        Returns:
            A copy of the SymbolLibrary instance.
        """
        sl = SymbolLibrary()
        sl.symbols = copy.deepcopy(self.symbols)
        return sl

    def add_symbol(self, symbol: str, symbol_type: str, precedence: int, np_fn: str):
        """
        Adds a symbol to the library. A symbol should have a type, precedence, and numpy function associated with it.
        Type "op" should be used for symbols operating on two operands, "fn" for symbols operating on one operand,
        "lit" for constants with a known value (such as pi or e), "const" for constants/parameters without a value that
        need to be optimized, and "var" for variables whose values are provided as input data.

        We recommend you use a single token of "const" type as using multiple might lead to more work, errors, and less
        readability.

        For example, look at the default_symbols function for the SymbolLibrary class.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.add_symbol("sin", "fn", 5, "np.sin({})")
            >>> library.add_symbol("C", "const", 5, "C[{}]")
            >>> library.add_symbol("X", "var", 5, "X[:, 0]")
            >>> library.add_symbol("pi", "lit", 5, "np.pi")

        Args:
            symbol: The symbol to be added to the library.
            symbol_type: The type of the symbol, one of "op" (operator), "fn" (function), "lit" (literal), "const" (constant), or "var" (variable).
            precedence: The precedence of the symbol, used to determine the order of operations.
            np_fn: A string representing the numpy function associated with this symbol.
        """
        self.symbols[symbol] = {
            "symbol": symbol,
            "type": symbol_type,
            "precedence": precedence,
            "np_fn": np_fn,
        }

    def remove_symbol(self, symbol: str):
        """
        Removes a symbol from the library.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> len(library.symbols)
            1
            >>> library.remove_symbol("x")
            >>> len(library.symbols)
            0

        Args:
            symbol: The symbol to be removed from the library.

        Raises:
            KeyError: If the symbol does not exist in the library.
        """
        del self.symbols[symbol]

    def get_type(self, symbol: str) -> str:
        """
        Retrieves the type of a symbol from the library.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.get_type("x")
            'var'

        Args:
            symbol: The symbol whose type is to be retrieved.

        Returns:
            The type of the symbol if it exists in the library, otherwise an empty string.
        """
        if symbol in self.symbols:
            return self.symbols[symbol]["type"]
        else:
            return ""

    def get_precedence(self, symbol: str) -> int:
        """
        Retrieves the precedence of the given symbol.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.get_precedence("x")
            0

        Args:
            symbol: The symbol whose precedence is to be retrieved.

        Returns:
            The precedence of the symbol if it exists in the library, otherwise -1.
        """
        if symbol in self.symbols:
            return self.symbols[symbol]["precedence"]
        else:
            return -1

    def get_np_fn(self, symbol: str) -> str:
        """
        Returns the numpy function corresponding to the given symbol.

        Examples:
            >>> library = SymbolLibrary()
            >>> library.add_symbol("x", "var", 0, "x")
            >>> library.get_np_fn("x")
            'x'

        Args:
            symbol: The symbol to look up.

        Returns:
            The numpy function corresponding to the given symbol, or an empty string if the symbol was not found.
        """
        if symbol in self.symbols:
            return self.symbols[symbol]["np_fn"]
        else:
            return ""

    @staticmethod
    def default_symbols(num_variables: int = 25) -> "SymbolLibrary":
        """
        Creates a SymbolLibrary instance populated with default mathematical symbols.

        This method adds a set of predefined symbols to a SymbolLibrary instance,
        representing common mathematical operations, functions, constants, and optional
        variables. The symbols include basic arithmetic operations, trigonometric and
        exponential functions, and mathematical constants like pi and e.

        If num_variables is greater than 0, it adds variables labeled 'X_0' to 'X_{num_variables-1}', each
         associated with a column in a data array X.

        Note: The variables in the default_symbols function are added in the predefined order,
        which is the same order as the columns in the data array X.

        Examples:
            >>> library = SymbolLibrary.default_symbols()
            >>> len(library.symbols)
            44

        Args:
            num_variables: The number of variables to add to the library (default is 25).

        Returns:
            A SymbolLibrary instance populated with default mathematical symbols.
        """
        sl = SymbolLibrary()
        sl.add_symbol("+", symbol_type="op", precedence=0, np_fn="{} = {} + {}")
        sl.add_symbol("-", symbol_type="op", precedence=0, np_fn="{} = {} - {}")
        sl.add_symbol("*", symbol_type="op", precedence=1, np_fn="{} = {} * {}")
        sl.add_symbol("/", symbol_type="op", precedence=1, np_fn="{} = {} / {}")
        sl.add_symbol("^", symbol_type="op", precedence=2, np_fn="{} = np.pow({},{})")
        sl.add_symbol("u-", symbol_type="fn", precedence=5, np_fn="{} = -{}")
        sl.add_symbol("sqrt", symbol_type="fn", precedence=5, np_fn="{} = np.sqrt({})")
        sl.add_symbol("sin", symbol_type="fn", precedence=5, np_fn="{} = np.sin({})")
        sl.add_symbol("cos", symbol_type="fn", precedence=5, np_fn="{} = np.cos({})")
        sl.add_symbol("exp", symbol_type="fn", precedence=5, np_fn="{} = np.exp({})")
        sl.add_symbol("ln", symbol_type="fn", precedence=5, np_fn="{} = np.log({})")
        sl.add_symbol("log", symbol_type="fn", precedence=5, np_fn="{} = np.log10({})")
        sl.add_symbol("^-1", symbol_type="fn", precedence=-1, np_fn="{} = 1/{}")
        sl.add_symbol("^2", symbol_type="fn", precedence=-1, np_fn="{} = {}**2")
        sl.add_symbol("^3", symbol_type="fn", precedence=-1, np_fn="{} = {}**3")
        sl.add_symbol("^4", symbol_type="fn", precedence=-1, np_fn="{} = {}**4")
        sl.add_symbol("^5", symbol_type="fn", precedence=-1, np_fn="{} = {}**5")
        sl.add_symbol("pi", symbol_type="lit", precedence=5, np_fn="np.full(X.shape[0], np.pi)")
        sl.add_symbol("e", symbol_type="lit", precedence=5, np_fn="np.full(X.shape[0], np.e)")
        sl.add_symbol("C", symbol_type="const", precedence=5, np_fn="np.full(X.shape[0], C[{}])")

        if num_variables > 0:
            for i in range(num_variables):
                sl.add_symbol(f"X_{i}", "var", 5, "X[:, {}]".format(i))

        return sl


"""
This module contains the ParameterEstimator class, which is used to estimate the parameters of an expression.
"""
from typing import Optional, List, Tuple

import numpy as np
from scipy.optimize import minimize

from SRToolkit.utils.symbol_library import SymbolLibrary
from SRToolkit.utils.expression_compiler import expr_to_error_function


class ParameterEstimator:
    def __init__(self, X: np.ndarray, y: np.ndarray, symbol_library: SymbolLibrary=SymbolLibrary.default_symbols(), **kwargs):
        """
        Initializes an instance of the ParameterEstimator class.

        Examples:
            >>> X = np.array([[1, 2], [8, 4], [5, 4], [7, 9], ])
            >>> y = np.array([3, 0, 3, 11])
            >>> pe = ParameterEstimator(X, y)
            >>> rmse, constants = pe.estimate_parameters(["C", "*", "X_1", "-", "X_0"])
            >>> print(rmse < 1e-6)
            True
            >>> print(1.99 < constants[0] < 2.01)
            True

        Args:
            X: The input data to be used in parameter estimation for variables. We assume that X is a 2D array
                with shape (n_samples, n_features).
            y: The target values to be used in parameter estimation.
            symbol_library: The symbol library to use. Defaults to SymbolLibrary.default_symbols().

        Keyword Arguments:
            method str: The method to be used for minimization. Currently, only "L-BFGS-B" is supported/tested. Default is "L-BFGS-B".
            tol float: The tolerance for termination. Default is 1e-6.
            gtol float: The tolerance for the gradient norm. Default is 1e-3.
            max_iter int: The maximum number of iterations. Default is 100.
            bounds List[float]: A list of two elements, specifying the lower and upper bounds for the constant values. Default is [-5, 5].
            initialization str: The method to use for initializing the constant values. Currently, only "random" and "mean" are supported. "random" creates a vector with random values
                                sampled within the bounds. "mean" creates a vector where all values are calculated as (lower_bound + upper_bound)/2. Default is "random".
            max_constants int: The maximum number of constants allowed in the expression. Default is 8.
            max_expr_length int: The maximum length of the expression. Default is -1 (no limit).

        Methods:
            estimate_parameters(expr: List[str]): Estimates the parameters of an expression by minimizing the error between the predicted and actual values.
        """
        self.symbol_library = symbol_library
        self.X = X
        self.y = y
        # self.stats = {"success": 0, "failure": 0, "steps": dict(), "num_constants": dict(), "failed_constants": dict()}

        self.estimation_settings = {
                "method": "L-BFGS-B",
                "tol": 1e-6,
                "gtol": 1e-3,
                "max_iter": 100,
                "bounds": [-5, 5],
                "initialization": "random", # random, mean
                "max_constants": 8,
                "max_expr_length": -1
        }

        if kwargs:
            self.estimation_settings.update(kwargs)

    def estimate_parameters(self, expr: List[str]) -> Tuple[float, np.ndarray]:
        """
        Estimates the parameters of an expression by minimizing the error between the predicted and actual values.

        Examples:
            >>> X = np.array([[1, 2], [8, 4], [5, 4], [7, 9], ])
            >>> y = np.array([3, 0, 3, 11])
            >>> pe = ParameterEstimator(X, y)
            >>> rmse, constants = pe.estimate_parameters(["C", "*", "X_1", "-", "X_0"])
            >>> print(rmse < 1e-6)
            True
            >>> print(1.99 < constants[0] < 2.01)
            True

        Args:
            expr: A list of strings representing the expression to be evaluated. The expression should include the
                  symbol 'C' for constants whose values need to be estimated.

        Returns:
            the root mean square error (RMSE) of the optimized expression.
            An array containing the optimized constant values.

        Notes:
            if the length of the expression exceeds the maximum allowed, NaN and an empty array are returned.
            If the number of constants in the expression exceeds the maximum allowed, NaN and an empty array are returned.
            If there are no constants in the expression, the RMSE is calculated directly without optimization.
        """
        num_constants = sum([1 for t in expr if self.symbol_library.get_type(t) == "const"])
        if 0 <= self.estimation_settings["max_constants"] < num_constants:
            return np.nan, np.array([])

        if 0 <= self.estimation_settings["max_expr_length"] < len(expr):
            return np.nan, np.array([])

        executable_error_fn = expr_to_error_function(expr, self.symbol_library)

        if num_constants == 0:
            rmse = executable_error_fn(self.X, np.array([]), self.y)
            return rmse, np.array([])
        else:
            return self._optimize_parameters(executable_error_fn, num_constants)

    def _optimize_parameters(self, executable_error_fn: callable, num_constants: int) -> Tuple[float, np.ndarray]:
        """
        Optimizes the parameters of a given expression by minimizing the root mean squared error between the predicted and actual values.

        Parameters
        ----------
        executable_error_fn : callable
            A function that takes in the input values, the constant values, and the target values and returns the root mean squared error.
        num_constants : int
            The number of constants in the expression.

        Returns
        -------
        float
            The root mean square error of the optimized expression.
        np.ndarray
            An array containing the optimized constant values.
        """
        if self.estimation_settings["initialization"] == "random":
            x0 = np.random.rand(num_constants) * (self.estimation_settings["bounds"][1] - self.estimation_settings["bounds"][0]) + self.estimation_settings["bounds"][0]
        else:
            x0 = np.array([np.mean(self.estimation_settings["bounds"]) for _ in range(num_constants)])

        res = minimize(lambda c: executable_error_fn(self.X, c, self.y), x0, method=self.estimation_settings["method"],
                       tol=self.estimation_settings["tol"],
                       options={
                           "maxiter": self.estimation_settings["max_iter"],
                           "gtol": self.estimation_settings["gtol"]
                                },
                       bounds=[(self.estimation_settings["bounds"][0], self.estimation_settings["bounds"][1]) for _ in range(num_constants)])

        # if res.success:
        #     self.stats["success"] += 1
        # else:
        #     self.stats["failure"] += 1
        #     if num_constants in self.stats["failed_constants"]:
        #         self.stats["failed_constants"][num_constants] += 1
        #     else:
        #         self.stats["failed_constants"][num_constants] = 1
        #
        # if res.nit in self.stats["steps"]:
        #     self.stats["steps"][res.nit] += 1
        # else:
        #     self.stats["steps"][res.nit] = 1
        #
        # if num_constants in self.stats["num_constants"]:
        #     self.stats["num_constants"][num_constants] += 1
        # else:
        #     self.stats["num_constants"][num_constants] = 1

        return res.fun, res.x





"""The module containing the `utils`.

The `utils` module provides a set of utilities used in the package and for expression compilation.

Modules:
    symbol_library: The module containing the symbol library data structure for managing symbols that can occur in expressions and their properties.
    expression_tree: The module containing the expression tree data structure and functions for transforming expressions into trees and back.
    expression_compiler: The module containing functions that transform expressions in the infix notation (represented as lists of tokens) to executable python functions.

"""

from .expression_tree import Node, tokens_to_tree, is_float
from .symbol_library import SymbolLibrary
from .expression_compiler import tree_to_function_rec, expr_to_executable_function, expr_to_error_function


__all__ = ["SymbolLibrary", "Node", "tokens_to_tree", "is_float", "tree_to_function_rec", "expr_to_executable_function",
           "expr_to_error_function"]


"""
The module containing the expression tree data structure and functions for transforming expressions into trees and back.
"""
from typing import List
import warnings

from SRToolkit.utils.symbol_library import SymbolLibrary


class Node:
    def __init__(self, symbol: str = None, right: "Node" = None, left: "Node" = None):
        """
        Initializes a Node object. We assume that nodes containing functions have only one child node, i.e. right is None.

        Examples:
            >>> node = Node("+", Node("x"), Node("1"))
            >>> len(node)
            3

        Args:
            symbol: The symbol string stored in this node.
            right: The right child of this node.
            left: The left child of this node.

        Methods:
            __len__(self):
                Returns the number of nodes in the tree rooted at this node.
            __str__(self):
                Returns a string representation of the tree rooted at this node.
            to_list(self, notation: str = "infix", symbol_library: SymbolLibrary = None):
                Returns a list representation of the tree rooted at this node.

        """
        self.symbol = symbol
        self.right = right
        self.left = left

    def __len__(self) -> int:
        """
        Returns the number of nodes in the tree rooted at this node.

        Examples:
            >>> node = Node("+", Node("x"), Node("1"))
            >>> len(node)
            3

        Returns:
            The number of nodes in the tree rooted at this node.
        """
        return (
            1
            + (len(self.left) if self.left is not None else 0)
            + (len(self.right) if self.right is not None else 0)
        )

    def __str__(self) -> str:
        """
        Returns a string representation of the tree rooted at this node.

        Examples:
            >>> node = Node("+", Node("x"), Node("1"))
            >>> str(node)
            'x+1'

        Returns:
            A string representation of the tree rooted at this node.
        """
        return "".join(self.to_list())

    def to_list(self, notation: str = "infix", symbol_library: SymbolLibrary = None) -> List[str]:
        """
        Transforms the tree rooted at this node into a list of tokens.

        Examples:
            >>> node = Node("+", Node("X_0"), Node("1"))
            >>> node.to_list(symbol_library=SymbolLibrary.default_symbols())
            ['1', '+', 'X_0']
            >>> node.to_list(notation="postfix")
            ['1', 'X_0', '+']
            >>> node.to_list(notation="prefix")
            ['+', '1', 'X_0']
            >>> node = Node("+", Node("*", Node("X_0"), Node("X_1")), Node("1"))
            >>> node.to_list(symbol_library=SymbolLibrary.default_symbols())
            ['1', '+', 'X_1', '*', 'X_0']
            >>> node.to_list(notation="infix")
            ['1', '+', '(', 'X_1', '*', 'X_0', ')']
            >>> node = Node("sin", None, Node("X_0"))
            >>> node.to_list(symbol_library=SymbolLibrary.default_symbols())
            ['sin', '(', 'X_0', ')']
            >>> node = Node("^2", None, Node("X_0"))
            >>> node.to_list(symbol_library=SymbolLibrary.default_symbols())
            ['X_0', '^2']
            >>> node.to_list()
            ['(', 'X_0', ')', '^2']

        Args:
            notation: The notation to use for the resulting list of tokens. One of "prefix", "postfix", or "infix".
            symbol_library: The symbol library to use when converting the tree. This library defines the properties of the symbols in the tree.

        Returns:
            A list of tokens representing the tree rooted at this node in the specified notation.

        Raises:
             Exception: If the notation is not one of "prefix", "postfix", or "infix" or if a symbol is not in the symbol library.

        Notes:
            If the notation is "infix" and the symbol library is not provided, then the resulting list of tokens may contain unnecessary parentheses or have other issues.
        """
        left = [] if self.left is None else self.left.to_list(notation, symbol_library)
        right = [] if self.right is None else self.right.to_list(notation, symbol_library)

        if notation == "prefix":
            return [self.symbol] + left + right

        elif notation == "postfix":
            return left + right + [self.symbol]

        elif notation == "infix" and symbol_library is None:
            warnings.warn("Symbol library not provided. Generated expression may contain unnecessary parentheses and"
                          " have other issues.")
            if self.left is None and self.right is None:
                return [self.symbol]
            if self.right is None and self.left is not None:
                if self.symbol[0] == "^":
                    return ["("] + left + [")", self.symbol]
                else:
                    return [self.symbol, "("] + left + [")"]
            else:
                if len(left) > 1:
                    left = ["("] + left + [")"]
                if len(right) > 1:
                    right = ["("] + right + [")"]
                return left + [self.symbol] + right

        elif notation == "infix":
            if is_float(self.symbol):
                return [self.symbol]
            if symbol_library.get_type(self.symbol) in ["var", "const", "lit"]:
                return [self.symbol]
            elif symbol_library.get_type(self.symbol) == "fn":
                if symbol_library.get_precedence(self.symbol) > 0:
                    return [self.symbol, "("] + left + [")"]
                else:
                    if len(left) > 1:
                        left = ["("] + left + [")"]
                    return left + [self.symbol]
            elif symbol_library.get_type(self.symbol) == "op":
                if not is_float(self.left.symbol) and -1 < symbol_library.get_precedence(self.left.symbol) <= symbol_library.get_precedence(self.symbol):
                    left = ["("] + left + [")"]
                if not is_float(self.right.symbol) and -1 < symbol_library.get_precedence(self.right.symbol) <= symbol_library.get_precedence(self.symbol):
                    right = ["("] + right + [")"]
                return left + [self.symbol] + right
            else:
                raise Exception(f"Invalid symbol type for symbol {self.symbol}.")
        else:
            raise Exception("Invalid notation selected. Use 'infix', 'prefix', 'postfix', or leave blank (defaults to 'infix').")


def is_float(element: any) -> bool:
    """
    Checks if a given element is a float.

    Examples:
        >>> is_float(1.0)
        True
        >>> is_float("1.0")
        True
        >>> is_float("1")
        True
        >>> is_float(None)
        False


    Args:
        element: The element to check.

    Returns:
        True if the element is a float, False otherwise.
    """
    if element is None:
        return False
    try:
        float(element)
        return True
    except ValueError:
        return False


def tokens_to_tree(tokens: List[str], sl: SymbolLibrary) -> Node:
    """
    Converts a list of tokens to a tree data structure. Throws an exception if the expression is invalid (check syntax
    and that all symbols are in the symbol library correctly defined).

    Examples:
        >>> tree = tokens_to_tree(["(", "x", "+", "y", ")"], SymbolLibrary.default_symbols())
        >>> len(tree)
        3

    Args:
        tokens: The list of tokens to convert.
        sl: The symbol library to use when parsing the tokens.

    Returns:
        The root of the expression tree data structure.

    Raises:
        Exception: If the expression is invalid. Usually this means that a symbol is not in the symbol library or that
                   there is a syntactic error in the expression.
    """
    num_tokens = len([t for t in tokens if t != "(" and t != ")"])
    expr_str = "".join(tokens)
    tokens = ["("] + tokens + [")"]
    operator_stack = []
    out_stack = []
    for token in tokens:
        if token == "(":
            operator_stack.append(token)
        elif sl.get_type(token) in ["var", "const", "lit"] or is_float(token):
            out_stack.append(Node(token))
        elif sl.get_type(token) == "fn":
            if token[0] == "^":
                out_stack.append(Node(token, left=out_stack.pop()))
            else:
                operator_stack.append(token)
        elif sl.get_type(token) == "op":
            while (
                len(operator_stack) > 0
                and operator_stack[-1] != "("
                and sl.get_precedence(operator_stack[-1]) > sl.get_precedence(token)
            ):
                if sl.get_type(operator_stack[-1]) == "fn":
                    out_stack.append(Node(operator_stack.pop(), left=out_stack.pop()))
                else:
                    out_stack.append(
                        Node(operator_stack.pop(), out_stack.pop(), out_stack.pop())
                    )
            operator_stack.append(token)
        else:
            if token != ")":
                raise Exception(f"Invalid symbol \"{token}\" in expression {expr_str}. Did you add token \"{token}\" to the symbol library?")

            while len(operator_stack) > 0 and operator_stack[-1] != "(":
                if sl.get_type(operator_stack[-1]) == "fn":
                    out_stack.append(Node(operator_stack.pop(), left=out_stack.pop()))
                else:
                    out_stack.append(
                        Node(operator_stack.pop(), out_stack.pop(), out_stack.pop())
                    )
            operator_stack.pop()
            if len(operator_stack) > 0 and sl.get_type(operator_stack[-1]) == "fn":
                out_stack.append(Node(operator_stack.pop(), left=out_stack.pop()))
    if len(out_stack[-1]) == num_tokens:
        return out_stack[-1]
    else:
        raise Exception(f"Error while parsing expression {expr_str}.")


if __name__ == '__main__':
    tree = tokens_to_tree(["(", "X_0", "+", "tan", "(", "X_1", "-", "5.2", ")", ")"], SymbolLibrary.default_symbols(num_variables=2))


import os
from setuptools import setup, find_packages


with open("README.md", "r") as fl:
    long_description = fl.read()


def parse_requirements(file):
    required_packages = []
    with open(os.path.join(os.path.dirname(__file__), file)) as req_file:
        for line in req_file:
            required_packages.append(line.strip())
    return required_packages


setup(name='symbolic-regression-toolkit',
      version='1.2.2',
      url='https://github.com/smeznar/SymbolicRegressionToolkit',
      author='Sebastian Mežnar',
      author_email='smeznar@gmail.com',
      license='GNU General Public License v3.0',
      keywords=["symbolic regression", "equation discovery", "parameter estimation", "machine learning"],
      description='Symbolic Regression/Equation Discovery Toolkit',
      long_description=long_description,
      long_description_content_type='text/markdown',
      py_modules=['SRToolkit'],
      packages=find_packages(),
      classifiers=['Intended Audience :: Information Technology',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python :: 3',
                   "Topic :: Scientific/Engineering",
                   'Topic :: Scientific/Engineering :: Artificial Intelligence',
                   "Topic :: Scientific/Engineering :: Mathematics"],

      install_requires=parse_requirements('requirements.txt'))


"""
This module contains data sets and benchmarks for symbolic regression. A dataset represents a single equation with specific data, constraints for evaluation, etc.
A benchmark is a collection of datasets. Our library provides the user with two modified version of popular equation discovery benchmarks. Specifically Feynman and Nguyen.

Modules:
    sr_dataset: The module containing the SRDataset class, which can be used to create a dataset and easily evaluate equation discovery approaches.
    sr_benchmark: The module containing the SRBenchmark class, which can be used to create a benchmark i.e. a collection of datasets.
"""

from .srdataset import SRDataset
from .srbenchmark import SRBenchmark

__all__ = ["SRDataset", "SRBenchmark"]


"""
This module contains classes and functions for evaluating symbolic regression approaches. Mainly it contains classes that
can be used for parameter estimation and evaluation of mathematical expressions on some dataset.

Modules:
    parameter_estimator: The module containing classes and functions for parameter estimation.
    sr_evaluator: The module containing classes and functions for expressions on some dataset.

"""

from .parameter_estimator import ParameterEstimator
from .sr_evaluator import SR_evaluator

__all__ = ["ParameterEstimator", "SR_evaluator"]


"""
This module contains functions that convert an expression in infix notation to an executable python function.
"""
from typing import List, Tuple

from SRToolkit.utils import Node, tokens_to_tree, is_float
from SRToolkit.utils.symbol_library import SymbolLibrary

# Generated functions are defined (through exec) here, so numpy needs to be imported
import numpy as np

def expr_to_executable_function(expr: List[str], symbol_library: SymbolLibrary=SymbolLibrary.default_symbols()) -> callable:
    """
    Converts an expression in infix notation to an executable function.

    Examples:
        >>> executable_fun = expr_to_executable_function(["X_0", "+", "1"])
        >>> executable_fun(np.array([[1], [2], [3], [4]]), np.array([]))
        array([2, 3, 4, 5])
        >>> executable_fun = expr_to_executable_function(["pi"])
        >>> executable_fun(np.array([[1], [2], [3], [4]]), np.array([1]))
        array([3.14159265, 3.14159265, 3.14159265, 3.14159265])
        >>> executable_fun = expr_to_executable_function(["C"])
        >>> executable_fun(np.array([[1], [2], [3], [4]]), np.array([1]))
        array([1, 1, 1, 1])

    Args:
        expr : The expression in infix notation.
        symbol_library : The symbol library to use. Defaults to SymbolLibrary.default_symbols().

    Returns:
        An executable function that takes in a 2D array of input values and a 1D array of constant values and returns the output of the expression.
    """
    tree = tokens_to_tree(expr, symbol_library)
    code, symbol, var_counter, const_counter = tree_to_function_rec(tree, symbol_library)

    fun_string = "def _executable_expression_(X, C):\n"
    for c in code:
        fun_string += "\t" + c + "\n"
    fun_string += "\treturn " + symbol

    exec(fun_string)
    return locals()["_executable_expression_"]


def expr_to_error_function(expr: List[str], symbol_library: SymbolLibrary=SymbolLibrary.default_symbols()) -> callable:
    """
    Converts an expression in infix notation to an executable function that returns the root mean squared error between
    the output of the expression and the target values.

    Examples:
        >>> executable_fun = expr_to_error_function(["X_0", "+", "1"])
        >>> executable_fun(np.array([[1], [2], [3], [4]]), np.array([]), np.array([2, 3, 4, 5]))
        0.0

    Args:
        expr : The expression in infix notation.
        symbol_library : The symbol library to use. Defaults to SymbolLibrary.default_symbols().

    Returns:
        An executable function that takes in a 2D array of input values `X`, a 1D array of constant values `C`, and a 1D array of target values `y`. It returns the root mean squared error between the output of the expression and the target values.
    """
    tree = tokens_to_tree(expr, symbol_library)
    code, symbol, var_counter, const_counter = tree_to_function_rec(tree, symbol_library)

    fun_string = "def _executable_expression_(X, C, y):\n"
    for c in code:
        fun_string += "\t" + c + "\n"
    fun_string += f"\treturn np.sqrt(np.mean(({symbol}-y)**2))"

    exec(fun_string)
    return locals()["_executable_expression_"]


def tree_to_function_rec(tree: Node, symbol_library: SymbolLibrary, var_counter: int=0, const_counter: int=0) -> Tuple[List[str], str, int, int]:
    """
    Recursively converts a parse tree into a string of Python code that can be executed to evaluate the expression
    represented by the tree.

    Args:
        tree: The root of the parse tree to convert.
        symbol_library: The symbol library to use when converting the tree. This library defines the properties of the symbols in the tree.
        var_counter: The number of variables encountered so far. This is used to create a unique variable name for each variable.
        const_counter: The number of constants encountered so far. This is used to select the correct constant value from the constant array.

    Returns:
        A list of strings, where each string contains a line of Python code to execute to evaluate the expression represented by the tree.
        The name of the variable that represents the output of the expression.
        The updated value of `var_counter`.
        The updated value of `const_counter`.

    Raises:
        Exception: If the parse tree contains an invalid symbol.

    Notes:
        This function is a helper function for `expr_to_executable_function` and similar and should not be called directly
        unless you want to customize the way the expression is defined. For examples, see the code of `expr_to_executable_function` and `expr_to_error_function` in this module.


    """
    if tree.left is None and tree.right is None:
        if symbol_library.get_type(tree.symbol) in ["var", "lit"]:
            return [], symbol_library.get_np_fn(tree.symbol), var_counter, const_counter
        elif symbol_library.get_type(tree.symbol) == "const":
            return [], symbol_library.get_np_fn(tree.symbol).format(const_counter), var_counter, const_counter + 1
        else:
            if is_float(tree.symbol):
                return [], tree.symbol, var_counter, const_counter
            else:
                raise Exception(f"Encountered invalid symbol {tree.symbol} while converting to function.")

    elif tree.left is not None and tree.right is None:
        code, symbol, var_counter, const_counter = tree_to_function_rec(tree.left, symbol_library, var_counter, const_counter)
        output_symbol = "y_{}".format(var_counter)
        code.append(symbol_library.get_np_fn(tree.symbol).format(output_symbol, symbol))
        return code, output_symbol, var_counter + 1, const_counter

    else:
        left_code, left_symbol, var_counter, const_counter = tree_to_function_rec(tree.left, symbol_library, var_counter, const_counter)
        right_code, right_symbol, var_counter, const_counter = tree_to_function_rec(tree.right, symbol_library, var_counter, const_counter)
        output_symbol = "y_{}".format(var_counter)
        code = left_code + right_code
        code.append(symbol_library.get_np_fn(tree.symbol).format(output_symbol, left_symbol, right_symbol))
        return code, output_symbol, var_counter + 1, const_counter


"""
This module contains the SR_evaluator class, which is used for evaluating symbolic regression approaches.
"""
from typing import Optional, List
import warnings

import numpy as np

from SRToolkit.utils.symbol_library import SymbolLibrary
from SRToolkit.evaluation.parameter_estimator import ParameterEstimator


class SR_evaluator:
    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        max_evaluations: int = -1,
        metadata: Optional[dict] = None,
        symbol_library: SymbolLibrary = SymbolLibrary.default_symbols(),
        **kwargs
    ):
        """
        Initializes an instance of the SR_evaluator class. This class is used for evaluating symbolic regression approaches.

        Examples:
            >>> X = np.array([[1, 2], [8, 4], [5, 4], [7, 9], ])
            >>> y = np.array([3, 0, 3, 11])
            >>> se = SR_evaluator(X, y)
            >>> rmse = se.evaluate_expr(["C", "*", "X_1", "-", "X_0"])
            >>> print(rmse < 1e-6)
            True


        Attributes:
            models: A dictionary containing the results of previously evaluated expressions.
            max_evaluations: The maximum number of expressions to evaluate.
            metadata: An optional dictionary containing metadata about this evaluation. This could include information such as the dataset used, the model used, seed, etc.
            symbol_library: The symbol library to use.
            total_expressions: The total number of expressions considered.
            parameter_estimator: An instance of the ParameterEstimator class used for parameter estimation.

        Args:
            X: The input data to be used in parameter estimation for variables. We assume that X is a 2D array with shape (n_samples, n_features).
            y: The target values to be used in parameter estimation.
            max_evaluations: The maximum number of expressions to evaluate. Default is -1, which means no limit.
            metadata: An optional dictionary containing metadata about this evaluation. This could include information such as the dataset used, the model used, seed, etc.
            symbol_library: The symbol library to use.

        Keyword Arguments:
            method str: The method to be used for minimization. Currently, only "L-BFGS-B" is supported/tested. Default is "L-BFGS-B".
            tol float: The tolerance for termination. Default is 1e-6.
            gtol float: The tolerance for the gradient norm. Default is 1e-3.
            max_iter int: The maximum number of iterations. Default is 100.
            bounds List[float]: A list of two elements, specifying the lower and upper bounds for the constant values. Default is [-5, 5].
            initialization str: The method to use for initializing the constant values. Currently, only "random" and "mean" are supported. "random" creates a vector with random values
                                sampled within the bounds. "mean" creates a vector where all values are calculated as (lower_bound + upper_bound)/2. Default is "random".
            max_constants int: The maximum number of constants allowed in the expression. Default is 8.
            max_expr_length int: The maximum length of the expression. Default is -1 (no limit).

        Methods:
            evaluate_expr(expr): Evaluates an expression in infix notation and stores the result in memory to prevent re-evaluation.
            get_results(top_k): Returns the results of the evaluation.
        """
        self.models = dict()
        self.metadata = metadata
        self.symbol_library = symbol_library
        self.max_evaluations = max_evaluations
        self.total_expressions = 0
        self.parameter_estimator = ParameterEstimator(
            X, y, symbol_library=symbol_library, **kwargs)

    def evaluate_expr(self, expr: List[str]) -> float:
        """
        Evaluates an expression in infix notation and stores the result in
        memory to prevent re-evaluation.

        Examples:
            >>> X = np.array([[1, 2], [8, 4], [5, 4], [7, 9], ])
            >>> y = np.array([3, 0, 3, 11])
            >>> se = SR_evaluator(X, y)
            >>> rmse = se.evaluate_expr(["C", "*", "X_1", "-", "X_0"])
            >>> print(rmse < 1e-6)
            True

        Args:
            expr: A list of strings representing the expression in infix notation.

        Returns:
            The root mean square error of the expression.

        Warnings:
            Maximum number of evaluations reached: If the maximum number of evaluations has been reached, a warning is printed and np.nan is returned.

        Notes:
            If the expression has already been evaluated, its stored value is returned instead of re-evaluating the expression.
            When the maximum number of evaluations has been reached, a warning is printed and np.nan is returned.
        """
        self.total_expressions += 1

        if 0 <= self.max_evaluations < self.total_expressions:
            warnings.warn(
                f"Maximum number of evaluations ({self.max_evaluations}) reached. Stopping evaluation.")
            return np.nan
        else:
            expr_str = "".join(expr)
            if expr_str in self.models:
                # print(f"Already evaluated {expr_str}")
                # print(self.models[expr_str])
                return self.models[expr_str]["rmse"]
            else:
                rmse, parameters = self.parameter_estimator.estimate_parameters(expr)
                self.models[expr_str] = {
                    "rmse": rmse,
                    "parameters": parameters,
                    "expr": expr,
                }
                return rmse

    # def evaluate_exprs(
    #     self, exprs: List[List[str]], num_processes: int = 1
    # ) -> List[float]:
    #     if num_processes > 1:
    #         pool = Pool(num_processes)
    #         results = pool.map(self.evaluate_expr, exprs)
    #         pool.close()
    #         for r in results:
    #             self.models
    #         return results
    #     else:
    #         return [self.evaluate_expr(expr) for expr in exprs]

    def get_results(self, top_k: int = 20, success_threshold: float = 1e-7) -> dict:
        """
        Returns the results of the equation discovery/symbolic regression process/evaluation.

        Examples:
            >>> X = np.array([[1, 2], [8, 4], [5, 4], [7, 9], ])
            >>> y = np.array([3, 0, 3, 11])
            >>> se = SR_evaluator(X, y)
            >>> rmse = se.evaluate_expr(["C", "*", "X_1", "-", "X_0"])
            >>> results = se.get_results(top_k=1)
            >>> print(results["num_evaluated"])
            1
            >>> print(results["total_expressions"])
            1
            >>> print(results["best_expr"])
            C*X_1-X_0
            >>> print(results["min_rmse"] < 1e-6)
            True
            >>> print(1.99 < results["results"][0]["parameters"][0] < 2.01)
            True

        Args:
            top_k: The number of top results to include in the output. If `top_k`
                is greater than the number of evaluated expressions, all
                evaluated expressions are included. If `top_k` is less than 0,
                all evaluated expressions are included.
            success_threshold: The threshold below which the evaluation is
                considered successful. Default is 1e-7.

        Returns:
            A dictionary containing the results of the equation discovery/symbolic regression process. The keys are:

                - "metadata" : The metadata provided in the constructor.
                - "min_rmse" : The minimum root mean squared error.
                - "best_expr" : The expression with the minimum root mean
                  squared error.
                - "num_evaluated" : The number of evaluated expressions.
                - "total_expressions" : The total number of expressions
                  considered.
                - "success" : Whether the evaluation was successful.
                - "results" : A list of dictionaries, where each dictionary
                  contains the root mean squared error, the expression, and the
                  estimated parameters of the expression. The list is sorted in
                  ascending order of the root mean squared error.
        """
        if top_k > len(self.models) or top_k < 0:
            top_k = len(self.models)

        models = list(self.models.values())
        best_indices = np.argsort([v["rmse"] for v in models])

        results = {
            "metadata": self.metadata,
            "min_rmse": models[best_indices[0]]["rmse"],
            "best_expr": "".join(models[best_indices[0]]["expr"]),
            "num_evaluated": len(models),
            "total_expressions": self.total_expressions,
            "results": list(),
        }

        # Determine success based on the predefined success threshold
        if success_threshold is not None and results["min_rmse"] < success_threshold:
            results["success"] = True
        else:
            results["success"] = False

        for i in best_indices[:top_k]:
            results["results"].append(models[i])

        return results
